@@ -21,6 +21,7 @@ pyexec_globals_utility_code = UtilityCode.load("PyExecGlobals", "Builtins.c")
 globals_utility_code = UtilityCode.load("Globals", "Builtins.c")
 
 builtin_utility_code = {
+    'StopAsyncIteration': UtilityCode.load_cached("StopAsyncIteration", "Coroutine.c"),
 }
 
 
@@ -94,27 +95,27 @@ builtin_function_table = [
                     is_strict_signature = True),
     BuiltinFunction('abs',        "f",    "f",     "fabsf",
                     is_strict_signature = True),
-    BuiltinFunction('abs',        None,    None,   "__Pyx_abs_int",
-                    utility_code = UtilityCode.load("abs_int", "Builtins.c"),
+    ] + list(
+        # uses getattr to get PyrexTypes.c_uint_type etc to allow easy iteration over a list
+        BuiltinFunction('abs',        None,    None,   "__Pyx_abs_{0}".format(t),
+                    utility_code = UtilityCode.load("abs_{0}".format(t), "Builtins.c"),
                     func_type = PyrexTypes.CFuncType(
-                        PyrexTypes.c_uint_type, [
-                            PyrexTypes.CFuncTypeArg("arg", PyrexTypes.c_int_type, None)
+                        getattr(PyrexTypes,"c_u{0}_type".format(t)), [
+                            PyrexTypes.CFuncTypeArg("arg", getattr(PyrexTypes,"c_{0}_type".format(t)), None)
                             ],
-                        is_strict_signature = True)),
-    BuiltinFunction('abs',        None,    None,   "__Pyx_abs_long",
-                    utility_code = UtilityCode.load("abs_long", "Builtins.c"),
+                        is_strict_signature = True, nogil=True))
+                            for t in ("int", "long", "longlong")
+             ) + list(
+        BuiltinFunction('abs',        None,    None,   "__Pyx_c_abs{0}".format(t.funcsuffix),
                     func_type = PyrexTypes.CFuncType(
-                        PyrexTypes.c_ulong_type, [
-                            PyrexTypes.CFuncTypeArg("arg", PyrexTypes.c_long_type, None)
+                        t.real_type, [
+                            PyrexTypes.CFuncTypeArg("arg", t, None)
                             ],
-                        is_strict_signature = True)),
-    BuiltinFunction('abs',        None,    None,   "__Pyx_abs_longlong",
-                    utility_code = UtilityCode.load("abs_longlong", "Builtins.c"),
-                    func_type = PyrexTypes.CFuncType(
-                        PyrexTypes.c_ulonglong_type, [
-                            PyrexTypes.CFuncTypeArg("arg", PyrexTypes.c_longlong_type, None)
-                        ],
-                        is_strict_signature = True)),
+                            is_strict_signature = True, nogil=True))
+                        for t in (PyrexTypes.c_float_complex_type,
+                                  PyrexTypes.c_double_complex_type,
+                                  PyrexTypes.c_longdouble_complex_type)
+                        ) + [
     BuiltinFunction('abs',        "O",    "O",     "PyNumber_Absolute"),
     #('all',       "",     "",      ""),
     #('any',       "",     "",      ""),
@@ -143,7 +144,8 @@ builtin_function_table = [
                     utility_code=getattr3_utility_code),
     BuiltinFunction('getattr',    "OO",   "O",     "__Pyx_GetAttr",
                     utility_code=getattr_utility_code),
-    BuiltinFunction('hasattr',    "OO",   "b",     "PyObject_HasAttr"),
+    BuiltinFunction('hasattr',    "OO",   "b",     "__Pyx_HasAttr",
+                    utility_code = UtilityCode.load("HasAttr", "Builtins.c")),
     BuiltinFunction('hash',       "O",    "h",     "PyObject_Hash"),
     #('hex',       "",     "",      ""),
     #('id',        "",     "",      ""),
@@ -319,12 +321,14 @@ builtin_types_table = [
                                     BuiltinMethod("clear",   "T",  "r", "PySet_Clear"),
                                     # discard() and remove() have a special treatment for unhashable values
 #                                    BuiltinMethod("discard", "TO", "r", "PySet_Discard"),
-                                    BuiltinMethod("update",     "TO", "r", "__Pyx_PySet_Update",
-                                                  utility_code=UtilityCode.load_cached("PySet_Update", "Builtins.c")),
+                                    # update is actually variadic (see Github issue #1645)
+#                                    BuiltinMethod("update",     "TO", "r", "__Pyx_PySet_Update",
+#                                                  utility_code=UtilityCode.load_cached("PySet_Update", "Builtins.c")),
                                     BuiltinMethod("add",     "TO", "r", "PySet_Add"),
                                     BuiltinMethod("pop",     "T",  "O", "PySet_Pop")]),
     ("frozenset", "PyFrozenSet_Type", []),
     ("Exception", "((PyTypeObject*)PyExc_Exception)[0]", []),
+    ("StopAsyncIteration", "((PyTypeObject*)__Pyx_PyExc_StopAsyncIteration)[0]", []),
 ]
 
 
@@ -380,6 +384,8 @@ def init_builtin_types():
             objstruct_cname = None
         elif name == 'Exception':
             objstruct_cname = "PyBaseExceptionObject"
+        elif name == 'StopAsyncIteration':
+            objstruct_cname = "PyBaseExceptionObject"
         else:
             objstruct_cname = 'Py%sObject' % name.capitalize()
         the_type = builtin_scope.declare_builtin_type(name, cname, utility, objstruct_cname)
@@ -406,11 +412,6 @@ def init_builtins():
         '__debug__', PyrexTypes.c_const_type(PyrexTypes.c_bint_type),
         pos=None, cname='(!Py_OptimizeFlag)', is_cdef=True)
 
-    entry = builtin_scope.declare_var(
-        'StopAsyncIteration', PyrexTypes.py_object_type,
-        pos=None, cname='__Pyx_PyExc_StopAsyncIteration')
-    entry.utility_code = UtilityCode.load_cached("StopAsyncIteration", "Coroutine.c")
-
     global list_type, tuple_type, dict_type, set_type, frozenset_type
     global bytes_type, str_type, unicode_type, basestring_type, slice_type
     global float_type, bool_type, type_type, complex_type, bytearray_type
