@@ -2,22 +2,31 @@
 #   Errors
 #
 
+from __future__ import absolute_import
+
+try:
+    from __builtin__ import basestring as any_string_type
+except ImportError:
+    any_string_type = (bytes, str)
+
 import sys
-from Cython.Utils import open_new_file
-import DebugFlags
-import Options
+
+from ..Utils import open_new_file
+from . import DebugFlags
+from . import Options
 
 
 class PyrexError(Exception):
     pass
 
+
 class PyrexWarning(Exception):
     pass
 
 
 def context(position):
     source = position[0]
-    assert not (isinstance(source, unicode) or isinstance(source, str)), (
+    assert not (isinstance(source, any_string_type)), (
         "Please replace filename strings with Scanning.FileSourceDescriptor instances %r" % source)
     try:
         F = source.get_lines()
@@ -48,10 +57,17 @@ class CompileError(PyrexError):
     def __init__(self, position = None, message = u""):
         self.position = position
         self.message_only = message
+        self.formatted_message = format_error(message, position)
         self.reported = False
     # Deprecated and withdrawn in 2.6:
     #   self.message = message
-        Exception.__init__(self, format_error(message, position))
+        Exception.__init__(self, self.formatted_message)
+        # Python Exception subclass pickling is broken,
+        # see http://bugs.python.org/issue1692335
+        self.args = (position, message)
+
+    def __str__(self):
+        return self.formatted_message
 
 class CompileWarning(PyrexWarning):
 
@@ -96,6 +112,9 @@ class CompilerCrash(CompileError):
                 message += u'\n'
             message += u'%s: %s' % (cause.__class__.__name__, cause)
         CompileError.__init__(self, pos, message)
+        # Python Exception subclass pickling is broken,
+        # see http://bugs.python.org/issue1692335
+        self.args = (pos, context, message, cause, stacktrace)
 
 class NoElementTreeInstalledException(PyrexError):
     """raised when the user enabled options.gdb_debug but no ElementTree
@@ -126,8 +145,8 @@ def close_listing_file():
         listing_file.close()
         listing_file = None
 
-def report_error(err):
-    if error_stack:
+def report_error(err, use_stack=True):
+    if error_stack and use_stack:
         error_stack[-1].append(err)
     else:
         global num_errors
@@ -147,12 +166,13 @@ def report_error(err):
             try: echo_file.write(line)
             except UnicodeEncodeError:
                 echo_file.write(line.encode('ASCII', 'replace'))
-        num_errors = num_errors + 1
+        num_errors += 1
         if Options.fast_fail:
-            raise AbortError, "fatal errors"
+            raise AbortError("fatal errors")
+
 
 def error(position, message):
-    #print "Errors.error:", repr(position), repr(message) ###
+    #print("Errors.error:", repr(position), repr(message)) ###
     if position is None:
         raise InternalError(message)
     err = CompileError(position, message)
@@ -160,7 +180,9 @@ def error(position, message):
     report_error(err)
     return err
 
-LEVEL=1 # warn about all errors level 1 or higher
+
+LEVEL = 1 # warn about all errors level 1 or higher
+
 
 def message(position, message, level=1):
     if level < LEVEL:
@@ -173,6 +195,7 @@ def message(position, message, level=1):
         echo_file.write(line)
     return warn
 
+
 def warning(position, message, level=0):
     if level < LEVEL:
         return
@@ -186,6 +209,7 @@ def warning(position, message, level=0):
         echo_file.write(line)
     return warn
 
+
 _warn_once_seen = {}
 def warn_once(position, message, level=0):
     if level < LEVEL or message in _warn_once_seen:
